from concurrent.futures import ThreadPoolExecutor # to donwload files concurrently
from typing import List
from urllib.request import urlopen # for requesting url
from functools import partial # required for progress bar
import os.path # for file path
import os, ssl # to fix ssl certificate verification
import typer # for CLI

# for progress bar
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)


# configuring/constructing progress bar
progress = Progress(
    TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
    BarColumn(bar_width=None),
    "[progress.percentage]{task.percentage:>3.1f}%",
    "•",
    DownloadColumn(),
    "•",
    TransferSpeedColumn(),
    "•",
    TimeRemainingColumn(),
)


def getFilename(response, url):
    # getting file name
    if response.info()["Content-type"] == "text/html":
        if url.endswith("/"):
            return url.split("/")[-2] + ".html"
        elif url.split(".")[-1] == "html":
            filename = url.split("/")[-1]
        else:
            return url.split("/")[-1] + ".html"

    else:
        return url.split("/")[-1]


def download_url(url: str, path: str):
    if (not os.environ.get('PYTHONHTTPSVERIFY', '') and getattr(ssl, '_create_unverified_context', None)):
        ssl._create_default_https_context = ssl._create_unverified_context # fixing ssl certification error

    with progress:
        progress.console.print(f"Requesting {url}")
        response = urlopen(url)

        # getting file name
        filename = getFilename(response=response, url=url)

        dest_path = os.path.join(path, filename) # creating file path
        filesize = int(response.info()["Content-length"]) # getting the size of the file

        # creating dir if it doesn't exist
        if not os.path.exists(path):
            os.mkdir(path)

        with open(dest_path, "wb") as dest_file:
            task = progress.add_task("downloading", filename=filename, total=filesize)

            try:
                for data in iter(partial(response.read, filesize), b""):
                    dest_file.write(data) # writing the data to the file

                    progress.update(task, advance=len(data)) # updating progress bar

            except OverflowError: # if file is too large then data will be written in parts(each of 20000 bytes)
                while True:
                    for data in iter(partial(response.read, 20000), b""):
                        dest_file.write(data) # writing the data to the file

                        progress.update(task, advance=len(data)) # updating progress bar
                    break

        progress.console.log(f"Downloaded [bold magenta]{dest_path}[/]")


def download(
        url: List[str] = typer.Option(..., help="URLs of the file to download."),
        path: str = typer.Option("./", help="Path where files will be downloaded.")
    ):
    """An URL file downloader which can download files over internet concurrently. Made in pure python."""

    with ThreadPoolExecutor(4) as pool:
        for urls in url:
            pool.submit(download_url, urls, path)


def main():
    typer.run(download)

main()